# -*- coding: utf-8 -*-

from noseapp_selenium.page_object.base import PageObject
from noseapp_selenium.page_object.base import WaitConfig
from noseapp_selenium.page_object.router import PageRouter


__all__ = (
    PageRouter,
    PageObject,
    WaitConfig,
)
